/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "flag.h"
#include "oi_wavelength.h"
#include "oi_vis.h"
#include "oi_target.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072

/* external int errno;*/
/*
    Purpose: produce calibrated visibilities, spectrophotometry and correlated
    fluxes as a function of wavelength, assuming that previously reduced data is
    available for target and calibrator.
  
    Usage::
    oirCalibrateVis visTag calTag [-sp spTag] [-calflux calFlux10] [-calspec fluxFile] [-cald calDiameter] [-nophot]

    Tag:    prefix pointing to reduced target .fits files e.g. Tag.redcal.fits
    calTag: prefix pointing to reduced visibility calibrator files
    spTag:  (optional) prefix pointing to separate spectorphotometric calibrator
            if omitted, calTag will be used.
    calflux: (optional) spcalibrator flux at 10 mu (Rayleigh-Jeans spectrum assumed).  
            If omitted, 1.0 Jy is used
    calspec: (optional) Name of FITS fdata file containing the calibrator flux (Jy) on EXACTLY the
             same wavelength points as the photometry/correlated flux data
    cald:   (optional) visibility calibrator diameter in milliarcsec.  If ommitted, 0.0mas assumed
    nophot: (optional)  If specified the .calcorr.fits output tables of correlated flux are
             computed using nophotometry option 
                target_calib_corr_flux = calibrator(Jy) * target_raw_corr_flux(adu)/ calib_raw_corr_flux(adu)
      default: use visibilities and photometry: 
                target_calib_corr_flux = targ_calib_total_photometry * targ_calib_visibility

    Output: 3 FitsFiles Tag.calvis.fits Tag.calphot.fits Tag.calcorr.fits

    containing:  
    I.  (a file in oi_vis format)
    1. a primary header (copied from Tag.corr.fits)
    2. an oi_wavelength table (copied from Tag.redcal.fits)
    3. an oi_target table (copied from Tag.redcal.fits if present)
    4. a oi_vis table with 1 row per input row in Tag: calibrated visibility
     
    II. an fdata file with detector table and 4 rows: 
    1.  Total photometry(Jy).
    2.  photometry (Jy) under mask
    3.  Estimated rms errors in (1.)
    4.  Estimated rms errors in (2.)

    III. (a file in oi_vis format)
    1.   a primary header (copied from Tag.corr.fits)
    2. an oi_wavelength table (copied from Tag.redcal.fits)
    3. an oi_target table (copied from Tag.redcal.fits if present)
    4. a oi_vis table with 1 row per input row in Tag: calibrated correlated flux


*/

void Usage(void)
/*----------------------------------------------------------------------- */
/*   Tells about usage of program and bails out                           */
/*----------------------------------------------------------------------- */
{
    fprintf(stderr, 
       "Usage: oirCalibrateVis visTag calTag [-sp spTag] [-calflux calFlux10] [-cald calDiameter] [-nophot] \n");
    fprintf(stderr, "Arguments:\n");
    fprintf(stderr, "  visTag: prefix pointing to reduced target .fits files e.g. visTag.redcal.fits\n");
    fprintf(stderr, "  calTag: prefix pointing to reduced visibility calibrator files\n");
    fprintf(stderr, "  -spTag: (optional) prefix pointing to separate spectorphotometric calibrator\n");
    fprintf(stderr, "  -calflux: (optional) spcalibrator flux at 10 mu.  Default = 1.Jy\n");
    fprintf(stderr, "  -calspec: (optional) FITS file containing calibrator spectrum. If specified\n");
    fprintf(stderr, "            this overrides calflux\n");
    fprintf(stderr, "  -cald: (optional) visibility calibrator diameter in milliarcsec.  Default=0.mas\n");
    fprintf(stderr, "  -nophot:(optional) compute correlated fluxes from raw data without using photometry data; \n");
    fprintf(stderr, "     default is to use photometry data \n");
    
    /*/exit(1);  bail out */
  }/* end Usage */

/*     J1(x)/x   */
float besseljx(float x) {
   float xx, y, f, theta, z;

   if (x >= 0.) xx = x; else xx = -x;
   y = xx/3.;

   if (y <= 1){
      y = y*y;  /* (x/3)^2 */
      z = 0.5 + 
      y * (-.56249985 + 
      y * (+.21093573 + 
      y * (-.03954289 + 
      y * (+.00443319 +
      y * (-.00031761 +
      y *.00001109)))));
   } else {
      y = 1./y;
      f = .79788456   +
      y * (.00000156  +
      y * (.01659667  +
      y * (.00017105  +
      y * (-.00249511 +
      y * (.00113653  +
      y * -.0020033)))));

      theta = xx -2.35619449 +
      y * (.12499612  +
      y * (.00005650  +
      y * (-.00637879 +
      y * (.00074348  +
      y * (.00079824  +
      y * -.00029166)))));

      z = f * cos(theta)/(xx * sqrt(xx));
   }
   return z;
}
 
int main (int argc, char** argv) {

   int i, j, ierr;
   long nRow, iRow ;
   int nx, nWave ;

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   int iKey ;


   imaging_fdata_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *detectorDef;
   imaging_fdata_row *inData;
   imaging_fdata_row *outData;

   double *calVisAmp;     /* instrumental visibility (amp)*/
   double *calVisAmpErr;  /* instrumental visibility (amp)*/
   double *calVisPhase;   /* instrumental vis (phase) */
   double *calVisPhaseErr;/* instrumental vis (phase) */
   float *calPhotArith;  /* Jy/adu/s/chan arithmetic mean of A+B */
   float *calPhotGeom;   /* Jy/adu/s/chan geometric mean of A,B */
   float *calPhotErr;    /* Jy/adu/s/chan*/
   float *calSpectrum;   /* Jy */
   float *targPhotMask;  /* spectrophotometric flux*/
   float *targPhotTotal;  /* spectrophotometric flux*/
   float *targPhotMaskErr;  /* spectrophotometric flux*/
   float *calVisCorr;    /* resolution correction for calibrator */
   double *calCorrFlux;  /* target correlated flux (calibrated) */
   double *calCorrFluxErr;/* target correlated flux (calibrated) error */
   float *temp;
   float *wave;
   float  maxPhot;
   double  maxCalCorrFlux;

   float baseline;
   float calUV;

   vis_table_def *vTable;
   vis_table_row *vRow;
   vis_table_row *oRow;
   vis_table_def *oTable;

   wavelength_table_def *wTable;
   wavelength_table_row *wRow;

   target_table_def *targTable;
   target_table_row *targRow;

   char spTag[130];
   char *arg;
   char history[80];

   char inFileName[130];
   char outFileName[130];
   char calSpecFileName[130];

   float w,x,y,z;

   int  doPhot;

   float calDiameter = 0.;
   float calFlux10   = 1.;

   char *cal      = ".redcal.fits";
   char *phot     = ".photometry.fits";
   char *outcal   = ".calvis.fits";
   char *outcorr  = ".calcorr.fits";
   char *outphot  = ".calphot.fits";
   char *corr     = ".corr.fits";
   char *bang     = "!";
   int  iSpTag    = 0;
   calSpecFileName[0] = 0;  /* default no calibrator spectrum specified */

   doPhot = 1; /*  default use photometry data and visibilities to calculated correlated fluxes */
   ierr = 0;


/*       
      parse command line
*/
   if (argc <3) {
      Usage(); /* must have at least 2 arguments */
      return 1;
   }
   for (i=3; i<argc; i++) {
      arg = argv[i];
      if (strcmp(arg, "-cald") == 0) { /* calibrator diameter */
         sscanf(argv[++i], "%f", &calDiameter);
      } else if (strcmp(arg, "-calflux") ==0) { /* cal flux at 10 mu */
         sscanf(argv[++i], "%f", &calFlux10);
      } else if (strcmp(arg, "-sp") == 0) { /* separate sp calibrator */
         strcpy(spTag, argv[++i]);
         iSpTag = 1;
      } else if (strcmp(arg, "-calspec") == 0) { /* calibrator flux curve */
         strcpy(calSpecFileName, argv[++i]);
      } else if (strcmp(arg, "-nophot") == 0) { /* dont use photometry to calculate correlated fluxes */
         doPhot = 0;
      } else {
         Usage(); /* unrecognized option */
      }
   }  /* parse arg */
/*
      open calibrator redcal file and get instrumental visibility 
      amp+phase
*/
   strcpy(inFileName, argv[2]);
   if (doPhot) strcat(inFileName, cal); else strcat(inFileName,corr);

   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open redcal file  %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
     get projected baseline from ESO keywords
     calculate main part of resolution correction
*/
   i = fits_read_key(inFile, TFLOAT,  "HIERARCH ESO ISS PBL12 START", &baseline,
       NULL, &ierr);
   calUV = baseline *  3.1415926535 *  3.1415926535 * calDiameter /
      (180. * 3600. *1000.);  /* B * pi^2 * calDia (radians) */
/*
     get wavelength info
*/
   wTable = open_wavelength_table(inFile, &ierr);
/*
     copy  wavelength table to output file; wavelengths in meters
*/
   wRow   = make_wavelength_table_row(wTable);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   nWave = nRow;
   nx = nWave;
   wave = (float *)malloc(nx * sizeof(float));
   for (iRow =1;iRow < nRow; iRow++) {
      i = read_wavelength_table(inFile, wTable, wRow, iRow, &ierr);
      wave[iRow-1] = wRow->wave;
      }
   kill_wavelength_table_def(wTable);
   kill_wavelength_table_row(wRow);

/*
 *       create arrays to hold visibility data
*/
   calVisAmp      = (double *)malloc(nx * sizeof(double))   ; 
   calVisAmpErr   = (double *)malloc(nx * sizeof(double))   ; 
   calVisPhase    = (double *)malloc(nx * sizeof(double))   ;
   calVisPhaseErr = (double *)malloc(nx * sizeof(double))   ;
   calVisCorr     = (float *)malloc(nx * sizeof(float))   ;
   calPhotArith   = (float *)malloc(nx * sizeof(float))   ;
   calPhotGeom    = (float *)malloc(nx * sizeof(float))   ;
   calPhotErr     = (float *)malloc(nx * sizeof(float))   ;
   calSpectrum    = (float *)malloc(nx * sizeof(float))   ;
   targPhotMask   = (float *)malloc(nx * sizeof(float));
   targPhotTotal  = (float *)malloc(nx * sizeof(float));
   targPhotMaskErr= (float *)malloc(nx * sizeof(float));
   calCorrFlux   = (double *)malloc(nx * sizeof(double));
   calCorrFluxErr= (double *)malloc(nx * sizeof(double));
   temp = (float *)malloc(nx * sizeof(float));

   memset(calVisAmp,      0, nx*sizeof(double));
   memset(calVisAmpErr,   0, nx*sizeof(double));
   memset(calVisPhase,    0, nx*sizeof(double));
   memset(calVisPhaseErr, 0, nx*sizeof(double));
   memset(calPhotArith,   0, nx*sizeof(float));
   memset(calPhotGeom,    0, nx*sizeof(float));
   memset(calPhotErr,     0, nx*sizeof(float));
   memset(targPhotMask,   0, nx*sizeof(float));
   memset(targPhotTotal,  0, nx*sizeof(float));
   memset(targPhotMaskErr,0, nx*sizeof(float));

/*
     correction for resolution
*/
   for (i=0; i< nx; i++)  calVisCorr[i] = 2. * besseljx(calUV /wave[i]); 

/*
      open input vis table section
*/
   vTable = open_vis_table(inFile, &ierr);
   if (ierr != 0) {
      printf("fits open cal flux table error %i\n", ierr);
      return 1;
   }
   vRow = make_vis_row(vTable);
/*
              read instrumental visibility and phase
*/
   i = read_vis_table_row(inFile, vTable, vRow, 1, &ierr);
   for (i=0;i<nx;i++) {
      calVisAmp[i]      = vRow->visAmp[i];
      calVisAmpErr[i]   = vRow->visAmpErr[i];
      calVisPhase[i]    = vRow->visPhi[i];
      calVisPhaseErr[i] = vRow->visPhiErr[i];
   }
/*
      close input vis table, also closes file
*/
   i = close_vis_table(inFile, vTable, &ierr);
   kill_vis_row(vRow);
/*
      correct for resolution 
*/
   for (i=0; i<nx;i++) {
      calVisAmp[i]      = calVisAmp[i] /calVisCorr[i] ;
      calVisAmpErr[i]   = calVisAmpErr[i]/calVisCorr[i];
   }
/*
     Calibrator spectrum is not given in a file.
     Calculate unit calibrator flux assuming Rayleigh-Jeans law
     note wave in meters not microns
*/
   if (calSpecFileName[0] == 0) {
      for (i=0; i< nx; i++) {
         calSpectrum[i] = calFlux10 * 1.e-10/(wave[i]*wave[i]);
      }
/*             calibrator spectrum given in FITS file */
   } else {
      fits_open_file (&inFile, calSpecFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("   fits open calSpec file  %s gave error %i\n", calSpecFileName,
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
      detectorDef = read_imaging_detector_table(inFile, &ierr);
      if (detectorDef->naxis[0][0] != nx) {
         printf("   calSpectrum in %s has %i pixels; Data has %i pixels\n", 
            calSpecFileName, detectorDef->naxis[0][0], nx);
         return 1;
      }
      idata       = open_imaging_fdata_table(inFile, &ierr);
      if (ierr != 0) {
         printf("   error opening imaging data in %s \n", calSpecFileName);
         return 1;
      }
      inData      = make_imaging_fdata_row(idata);

      i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr);
      memcpy(calSpectrum, inData->data[0], nx*sizeof(float));
      i = close_imaging_fdata_table(inFile, idata, &ierr);
      kill_imaging_fdata_row(inData);
   }
/*              
 *                 ALL OF THE FOLLOWING PHOTOMETRY STUFF
 *                 IS ONLY IF doPhot = 1; i.e. there is
 *                 useful photometry
*/
   if (doPhot) {
/*
      get sp calibrator photometry data;
      take sum of TotalA + TotalB (rows 1&2)
      errors from rows 7&8
*/
   if (iSpTag == 0) strcpy(inFileName, argv[2]);
      else strcpy(inFileName, spTag);
   strcat(inFileName, phot);
   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open cal phot on %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
   idata          = open_imaging_fdata_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error reading cal phot data table %i",ierr);
      return 1;
   }
   inData = make_imaging_fdata_row(idata);
   i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr);
   for (i=0;i<nx;i++) {
      calPhotArith[i] = inData->data[0][i];
      calPhotGeom[i] = inData->data[0][i];
   }
   i = read_imaging_fdata_table(inFile, idata, inData, 2, &ierr);
   for (i=0;i<nx;i++) {
      calPhotArith[i] = 0.5*(calPhotArith[i] + inData->data[0][i]);
      calPhotGeom[i] = calPhotGeom[i] * inData->data[0][i];
      if (calPhotGeom[i] > 0.) calPhotGeom[i] = sqrt(calPhotGeom[i]);
         else calPhotGeom[i] = 0.;
   }
   i = read_imaging_fdata_table(inFile, idata, inData, 7, &ierr);
   for (i=0;i<nx;i++) calPhotErr[i] = inData->data[0][i]*inData->data[0][i];
   i = read_imaging_fdata_table(inFile, idata, inData, 8, &ierr);
   for (i=0;i<nx;i++) calPhotErr[i] = sqrt(calPhotErr[i] +
      inData->data[0][i]*inData->data[0][i]);
/*
      kill photometry table structure/close file
*/
   close_imaging_fdata_table(inFile, idata, &ierr);

/*
      open target input and output photometry data sets
*/
   
   strcpy(inFileName, argv[1]);
   strcat(inFileName, phot);

   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open source phot on %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }

/*   
      create output photometry file
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[1]);
   strcat(outFileName, outphot);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",
         outFileName, ierr);
      return 1;
   }
   i = fits_copy_header (inFile, outFile, &ierr);

/*
      record input parameters as HISTORY
*/
   snprintf(history,80, "oirCalibrateVis CalFlux10=%f", calFlux10);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history,80, "oirCalibrateVis CalDiameter=%f", calDiameter);
   i = fits_write_history(outFile, history, &ierr);
/*
      get detector info
*/
   detectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error reading detector table %i",ierr);
      return 1;
   }
   idata          = open_imaging_fdata_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error reading data table %i",ierr);
      return 1;
   }

   odata = make_imaging_fdata_table_def (outFile, 
      detectorDef, 1, &ierr); 
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
      }
   outData = make_imaging_fdata_row(odata);

/*
      read rows 1+2 and 6 (AB photometry total/mask)
      read rows 7+8 and 12 (AB photometry total/mask Errors)
      divide measured mask/total counts/s * calibrator
      and output.  Dont divide by values less than 1% of
      peak calibrator
*/
   maxPhot = -1.e9;
   for (i=0;i<nx;i++) if (calPhotArith[i] > maxPhot) maxPhot=calPhotArith[i];
   maxPhot = 0.01* maxPhot;
   for (i=0;i<nx;i++) if (calPhotArith[i] < maxPhot) calPhotArith[i] = maxPhot;
   maxPhot = -1.e9;
   for (i=0;i<nx;i++) if (calPhotGeom[i] > maxPhot) maxPhot=calPhotGeom[i];
   maxPhot = 0.01* maxPhot;
   for (i=0;i<nx;i++) if (calPhotGeom[i] < maxPhot) calPhotGeom[i] = maxPhot;

/*       
                      total photometry in Jy
*/
   i = read_imaging_fdata_table(inFile, idata, inData, 1, &ierr); /* A */
   for (i=0;i<nx;i++) outData->data[0][i] = inData->data[0][i];
   i = read_imaging_fdata_table(inFile, idata, inData, 2, &ierr); /* B */
   for (i=0;i<nx;i++) outData->data[0][i] = 0.5*(outData->data[0][i] + /* (A+B)*Jy/Cal(A+B) */
      inData->data[0][i])* calSpectrum[i]/calPhotArith[i];
   memcpy(targPhotTotal, outData->data[0], nx*sizeof(float)); /* save total target photometry */
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);

/*       
                      total photometry errors
*/
   i = read_imaging_fdata_table(inFile, idata, inData, 7, &ierr);
   for (i=0;i<nx;i++)temp[i] = inData->data[0][i]*inData->data[0][i];
   i = read_imaging_fdata_table(inFile, idata, inData, 8, &ierr);
   for (i=0;i<nx;i++)temp[i] = sqrt(temp[i]+inData->data[0][i]*inData->data[0][i]);
   for (i=0;i<nx;i++) {
      x = outData->data[0][i]; /* calFlux * targPhot/calPhot */
      y = calSpectrum[i] * temp[i];  /* targPhotErr */
      z = calPhotErr[i];
      outData->data[0][i] = sqrt(y*y + x*x*z*z)/calPhotArith[i];
   }
   i = write_imaging_fdata_table(outFile, odata, outData, 3, &ierr);

/*
                      mask photometry 
                      use geometric mean of total photometry to
                      convert to Jy
*/
   i = read_imaging_fdata_table(inFile, idata, inData, 6, &ierr); /* geom mean mask */
   memcpy(targPhotMask, inData->data[0], nx * sizeof(float));
   for (i=0;i<nx;i++) outData->data[0][i] = 
      calSpectrum[i] * inData->data[0][i]/calPhotGeom[i];
   i = write_imaging_fdata_table(outFile, odata, outData, 2, &ierr);

/*       
                      mask photometry errors
*/
   i = read_imaging_fdata_table(inFile, idata, inData, 12, &ierr);
   memcpy(targPhotMaskErr, inData->data[0], nx * sizeof(float));
   for (i=0;i<nx;i++) {
      x = outData->data[0][i]; /* targMask/calPhot */
      y = calSpectrum[i] * targPhotMaskErr[i];  /* targPhotErr */
      z = calPhotErr[i];
      outData->data[0][i] = sqrt(y*y + z*z*x*x)/calPhotGeom[i];
   }
   i = write_imaging_fdata_table(outFile, odata, outData, 4, &ierr);

/*
                      shut down target photometry files
*/
   close_imaging_fdata_table(inFile, idata, &ierr);
   close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(inData);
   kill_imaging_fdata_row(outData);
/*
      open input visibility data, output vis data
      copy headers and wavelength data
      read it row by row,
      divide by instrumental vis and masked photometry and output it
*/
/*
      open input and output vis data sets
*/
   
   strcpy(inFileName, argv[1]);
   strcat(inFileName, corr);

   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open source corr file on %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }

   strcpy(outFileName, bang);
   strcat(outFileName, argv[1]);
   strcat(outFileName, outcal);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening cal output %s %i\n",
         outFileName, ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }

   i = fits_copy_header (inFile, outFile, &ierr);
/*
      record input parameters as HISTORY
*/
   snprintf(history,80, "oirCalibrateVis CalFlux10=%f", calFlux10);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history,80, "oirCalibrateVis CalDiameter=%f", calDiameter);
   i = fits_write_history(outFile, history, &ierr);

   wTable = open_wavelength_table(inFile, &ierr);
/*
     copy  wavelength table to output file
*/
   i = fits_copy_hdu(inFile, outFile, 0, &ierr);
   kill_wavelength_table_def(wTable);

/*
 *    try to copy target table (if it is there)
*/

   targTable = open_target_table(inFile, &ierr);
   if (ierr) {
      printf("   no TARGET table in input file \n" );
      printf("   please ignore previous OI_TARGET error message 301\n");
      printf("   I am processing data without target information.\n");
      ierr = 0;
   } else {
      i = fits_copy_hdu(inFile, outFile, 0, &ierr);
      targRow   = make_target_table_row(targTable);
      i = read_target_table(inFile, targTable, targRow, 1, &ierr);
      i = write_target_table(outFile, targTable, targRow, 1, &ierr);
      kill_target_table_def(targTable);
      kill_target_table_row(targRow);
   }
/*
      open input and output data tables
*/
   vTable = open_vis_table(inFile, &ierr);
   if (ierr != 0) {
      printf("fits open target flux table error %i\n", ierr);
      return 1;
   }
   vRow = make_vis_row(vTable);

   oTable = make_vis_table_def(outFile, nx, "1776-07-04", "VLTI",
      "MIDI/PRISM", &ierr);
   oRow   = make_vis_row(oTable);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
/*
      truncate target photometry at 1% of peak
*/
   maxPhot = -1.e9;
   for (i=0;i<nx;i++) if (targPhotMask[i] > maxPhot) maxPhot=targPhotMask[i];
   maxPhot = 0.01* maxPhot;
   for (i=0;i<nx;i++) if (targPhotMask[i] < maxPhot) 
      targPhotMask[i] = maxPhot;
   for (iRow =1; iRow<=nRow; iRow++) {
      i = read_vis_table_row(inFile, vTable, vRow, 1, &ierr);
/*
      copy non data entries
*/
      oRow->targetId   = vRow->targetId;
      oRow->time       = vRow->time;
      oRow->mjd        = vRow->mjd;
      oRow->intTime    = vRow->intTime;
      oRow->ucoord     = vRow->ucoord;
      oRow->vcoord     = vRow->vcoord;
      oRow->station[0] = vRow->station[0];
      oRow->station[1] = vRow->station[1];
      memcpy(oRow->flag, vRow->flag, nx * sizeof(char));

/*
      divide target correlated flux by target visibility
      and then by instrumental visibility;
      error is in quotient usual logarithmic rms
*/
      for (i=0; i<nx; i++) {
         if ((targPhotMask[i]) > 0 && (calVisAmp[i]>0) && 
            (vRow->visAmp[i] > 0)) {
            x = vRow->visAmp[i]/(targPhotMask[i]*calVisAmp[i]);
            oRow->visAmp[i] = x;
            y = targPhotMaskErr[i]/targPhotMask[i];
            z = calVisAmpErr[i]/calVisAmp[i];
            w = vRow->visAmpErr[i]/vRow->visAmp[i];
            oRow->visAmpErr[i] = x * sqrt(y*y + z*z + w*w);
         } else {
            oRow->visAmp[i]    = 0.;
            oRow->visAmpErr[i] = 0.;
         }
         oRow->visPhi[i]= vRow->visPhi[i] - calVisPhase[i];
         if (oRow->visPhi[i] < -180.) oRow->visPhi[i] = 
            oRow->visPhi[i]+360.;
         oRow->visPhiErr[i] = sqrt(vRow->visPhiErr[i]*vRow->visPhiErr[i] +
            calVisPhaseErr[i]*calVisPhaseErr[i]);
      }
      i = write_vis_table_row(outFile, oTable, oRow, iRow, &ierr);
   }
   close_vis_table(outFile, oTable, &ierr);
   close_vis_table(inFile,  vTable, &ierr);
   kill_vis_row(oRow);

   }      /*   WE DID THIS BECAUSE doPhot was TRUE */
/*
      This section writes calibrated correlated flux according to two options:

      1. [default] correlated flux is calibrated visibility (which includes target photometry)
         times target TOTAL photometry.

      2. if -nophot is specified (crappy photometry), take a shortcut and produce
         calibrator flux (Jy) * raw target correlated flux/ raw calibrator correlated flux 

*/
/*
      if option (2) get calibrator raw correlated flux
*/
   if (doPhot == 0) {
      strcpy(inFileName, argv[2]);
      strcat(inFileName, corr);

      fits_open_file (&inFile, inFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("fits open calibrator corr file on %s gave error %i\n", inFileName,
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
/*
      open input data tables; read first line into data arrays
*/
      vTable = open_vis_table(inFile, &ierr);
      if (ierr != 0) {
         printf("fits open calibrator corr flux table error %i\n", ierr);
         return 1;
      }
      vRow = make_vis_row(vTable);
      i = read_vis_table_row(inFile, vTable, vRow, 1, &ierr);
      memcpy(calCorrFlux,    vRow->visAmp,    nx*sizeof(double));
      memcpy(calCorrFluxErr, vRow->visAmpErr, nx*sizeof(double));
      close_vis_table(inFile,  vTable, &ierr);
/*
     truncate raw calibrator corr flux
*/
      maxCalCorrFlux = -1.e9;
      for (i=0;i<nx;i++) if (calCorrFlux[i] > maxCalCorrFlux) maxCalCorrFlux=calCorrFlux[i];
      maxCalCorrFlux = 0.01* maxCalCorrFlux;

/*
      open input and output target corr flux data sets
*/
   
      strcpy(inFileName, argv[1]);
      strcat(inFileName, corr);

      fits_open_file (&inFile, inFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("fits open source corr file on %s gave error %i\n", inFileName,
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
   } else { /* end of doPhot == 0  start of doPhot == 1 */
      strcpy(inFileName, argv[1]);
      strcat(inFileName, outcal); /* get ready to reread calibrated visibilities */

      fits_open_file (&inFile, inFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("fits open source calvis file on %s gave error %i\n", inFileName,
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
   }  /* end of doPhot == 1 */
   strcpy(outFileName, bang);
   strcat(outFileName, argv[1]);
   strcat(outFileName, outcorr);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening calcorr output %s %i\n",
         outFileName, ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }

   i = fits_copy_header (inFile, outFile, &ierr);
/*
      record input parameters as HISTORY
*/
   snprintf(history,80, "oirCalibrateVis CalFlux10=%f", calFlux10);
   i = fits_write_history(outFile, history, &ierr);
   snprintf(history,80, "oirCalibrateVis CalDiameter=%f", calDiameter);
   i = fits_write_history(outFile, history, &ierr);

   wTable = open_wavelength_table(inFile, &ierr);
/*
     copy  wavelength table to output file
*/
   i = fits_copy_hdu(inFile, outFile, 0, &ierr);
   kill_wavelength_table_def(wTable);

/*
 *    try to copy target table (if it is there)
*/

   targTable = open_target_table(inFile, &ierr);
   if (ierr) {
      printf("   no TARGET table in input file \n" );
      printf("   please ignore previous OI_TARGET error message 301\n");
      printf("   I am processing data without target information.\n");
      ierr = 0;
   } else {
      i = fits_copy_hdu(inFile, outFile, 0, &ierr);
      targRow   = make_target_table_row(targTable);
      i = read_target_table(inFile, targTable, targRow, 1, &ierr);
      i = write_target_table(outFile, targTable, targRow, 1, &ierr);
      kill_target_table_def(targTable);
      kill_target_table_row(targRow);
   }
/*
      open input and output data tables
*/
   vTable = open_vis_table(inFile, &ierr);
   if (ierr != 0) {
      printf("in calCorr code fits open input flux table error %i\n", ierr);
      return 1;
   }

   oTable = make_vis_table_def(outFile, nx, "1776-07-04", "VLTI",
      "MIDI/PRISM", &ierr);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   for (iRow =1; iRow<=nRow; iRow++) {
      i = read_vis_table_row(inFile, vTable, vRow, 1, &ierr);

/*
      target correlated flux is calibrator nominal flux corrected
      for extension * target raw correlated flux / calibrator raw correlated flux
      error is in quotient usual logarithmic rms
*/
      if (doPhot == 1) {  /* default calCorr from visibility+photometry */
         for (i=0; i<nx; i++) {
            vRow->visAmp[i]    = vRow->visAmp[i]   *targPhotTotal[i];
            vRow->visAmpErr[i] = vRow->visAmpErr[i]*targPhotTotal[i]; /* assume error is in visibility */
            vRow->visPhi[i]= vRow->visPhi[i] - calVisPhase[i];
            if (vRow->visPhi[i] < -180.) vRow->visPhi[i] = 
               vRow->visPhi[i]+360.;
            if (vRow->visPhi[i] > 180.) vRow->visPhi[i] = 
               vRow->visPhi[i]-360.;
            vRow->visPhiErr[i] = sqrt(vRow->visPhiErr[i]*vRow->visPhiErr[i] +
               calVisPhaseErr[i]*calVisPhaseErr[i]);
         }
      } else {            /* optional calCorr from raw correlated fluxes */
         for (i=0; i<nx; i++) {
            if ((calCorrFlux[i]) > 0 && (vRow->visAmp[i] > 0)) {
               x = vRow->visAmp[i]*calSpectrum[i]*calVisCorr[i]/calCorrFlux[i];
               y = calCorrFluxErr[i]/calCorrFlux[i];
               w = vRow->visAmpErr[i]/vRow->visAmp[i];
               vRow->visAmp[i]    = x;
               vRow->visAmpErr[i] = x * sqrt(y*y + w*w);
            } else {
               vRow->visAmp[i]    = 0.;
               vRow->visAmpErr[i] = 0.;
            }
            vRow->visPhi[i]= vRow->visPhi[i] - calVisPhase[i];
            if (vRow->visPhi[i] < -180.) vRow->visPhi[i] = 
               vRow->visPhi[i]+360.;
            if (vRow->visPhi[i] > 180.) vRow->visPhi[i] = 
               vRow->visPhi[i]-360.;
            vRow->visPhiErr[i] = sqrt(vRow->visPhiErr[i]*vRow->visPhiErr[i] +
               calVisPhaseErr[i]*calVisPhaseErr[i]);
         }
      }
      i = write_vis_table_row(outFile, oTable, vRow, iRow, &ierr);
   }
   close_vis_table(outFile, oTable, &ierr);
   close_vis_table(inFile,  vTable, &ierr);
   kill_vis_row(vRow);

   free(calVisAmp);
   free(calVisPhase);
   free(calPhotArith);
   free(calPhotGeom);
   free(calVisCorr);
   free(calVisAmpErr);
   free(calVisPhaseErr);
   free(calPhotErr);
   free(calSpectrum);
   free(temp);
   free(targPhotMask);
   free(targPhotTotal);
   free(targPhotMaskErr);
   free(calCorrFlux);
   free(calCorrFluxErr);
   free(wave);
   return 0;
}
